library(caret)
library(ggplot2)

data = read.csv("communities.csv")

# ----1.----

columns_to_scale <- setdiff(names(data), "ViolentCrimesPerPop")

scaler <- preProcess(data[, columns_to_scale])

data_scaled <- data
data_scaled[, columns_to_scale] <- predict(scaler, data[, columns_to_scale])

X <- data_scaled[, columns_to_scale]
y <- data_scaled$ViolentCrimesPerPop

covariance_matrix <- cov(X)

eigen_decomposition <- eigen(covariance_matrix)

eigen_values <- eigen_decomposition$values

variance <- eigen_values / sum(eigen_values)

which(cumsum(variance) >= 0.95)[1]

variance[1]
variance[2]

# ----2.----

principle_components <- prcomp(X)

U <- principle_components$rotation

principle_components$x[, 1]

plot(U[, 1],
     main = "PC1",
     col = "blue",
     ylab = "Magnitude")

ordered_abs <- order(abs(U[, 1]), decreasing = TRUE)

names(U[, 1])[ordered_abs[1:5]]

pc_scores <- data.frame(
  PC1 = principle_components$x[, 1],
  PC2 = principle_components$x[, 2],
  ViolentCrimesPerPop = data_scaled$ViolentCrimesPerPop
)

ggplot(pc_scores, aes(x = PC1, y = PC2, color = ViolentCrimesPerPop)) +
  geom_point() +
  labs(title = "PC Scores", x = "PC1", y = "PC2") +
  scale_color_gradient(low = "blue", high = "red") +
  theme_minimal()

# ----3.----

n = dim(data)[1]
set.seed(12345)
id = sample(1:n, floor(n * 0.5))
train = data[id, ]
test = data[-id, ]

scaler3 <- preProcess(train[, columns_to_scale])

train_scaled <- train
train_scaled[, columns_to_scale] <- predict(scaler3, train[, columns_to_scale])

test_scaled <- test
test_scaled[, columns_to_scale] <- predict(scaler3, test[, columns_to_scale])

X_train <- train_scaled[, columns_to_scale]
y_train <- train_scaled$ViolentCrimesPerPop

X_test <- test_scaled[, columns_to_scale]
y_test <- test_scaled$ViolentCrimesPerPop

model <- lm(ViolentCrimesPerPop ~ . - 1, data = train_scaled)

summary(model)

MSE <- function(y, y_hat) {
  return(mean((y - y_hat) ^ 2))
}

y_hat_train <- predict(model, X_train)
y_hat_test <- predict(model, X_test)

MSE(y_train, y_hat_train)
MSE(y_test, y_hat_test)

# ----4.----

train_errors <- c()
test_errors <- c()

cost_function <- function(theta) {
  train_error <- MSE(y_train, as.matrix(X_train) %*% theta)
  test_error <- MSE(y_test, as.matrix(X_test) %*% theta)
  
  .GlobalEnv$train_errors <- c(train_errors, train_error)
  .GlobalEnv$test_errors <- c(test_errors, test_error)
  
  return(train_error)
}

res <- optim(par = rep(0, ncol(X_train)),
             fn = cost_function,
             method = "BFGS",)

range <- c(train_errors[500:length(train_errors)], test_errors[500:length(test_errors)])

plot(
  train_errors,
  type = "l",
  col = "blue",
  ylab = "Error",
  xlab = "Iteration",
  main = "Training and Test Errors over Iterations",
  lty = 1,
  ylim = c(min(range), max(range))
)

points(test_errors,
       type = "l",
       col = "red",
       lty = 2)

legend(
  "topright",
  legend = c("Training Error", "Test Error"),
  col = c("blue", "red"),
  lty = c(1, 2)
)

early_stopping <- which.min(test_errors[500:length(test_errors)]) + 500
early_stopping

test_errors[early_stopping]

optimal_theta <- res$par
optimal_train_mse <- res$value
optimal_train_mse

optimal_test_mse <- MSE(y_test, as.matrix(X_test) %*% optimal_theta)
optimal_test_mse
